import curses
import textwrap

from dataclasses import dataclass
from typing import TYPE_CHECKING, Optional

from mcp_server_webcrawl.interactive.ui import DocumentMode, ThemeDefinition, ViewBounds
from mcp_server_webcrawl.interactive.views.base import BaseCursesView, CONTENT_MARGIN
from mcp_server_webcrawl.interactive.highlights import HighlightProcessor, HighlightSpan
from mcp_server_webcrawl.models.resources import ResourceResult
from mcp_server_webcrawl.interactive.ui import safe_addstr
if TYPE_CHECKING:
    from mcp_server_webcrawl.interactive.session import InteractiveSession

DOCUMENT_MODE_NEXT: dict[DocumentMode, DocumentMode] = {
    DocumentMode.MARKDOWN: DocumentMode.RAW,
    DocumentMode.RAW: DocumentMode.HEADERS,
    DocumentMode.HEADERS: DocumentMode.MARKDOWN
}

@dataclass
class DocumentLineData:
    """
    Container for processed document line data with highlights.
    """
    original_line: str
    clean_text: str
    highlights: list[HighlightSpan]


class SearchDocumentView(BaseCursesView):
    """
    Document viewer with markdown/raw/headers modes, scrolling support, and search highlighting.
    """

    def __init__(self, session: 'InteractiveSession'):
        """
        Initialize the document view.
        
        Args:
            session: The interactive session instance
        """
        super().__init__(session)
        self.__document: Optional[ResourceResult] = None
        self.__scroll_offset: int = 0
        self.__document_mode: DocumentMode = DocumentMode.MARKDOWN
        self.__cached_content_lines: Optional[list[str]] = None
        self.__cached_mode: Optional[DocumentMode] = None
        self.__cached_query: Optional[str] = None
        self.__search_terms: list[str] = []

    @property
    def document_mode(self) -> DocumentMode:
        return self.__document_mode

    @property
    def scroll_offset(self) -> int:
        return self.__scroll_offset

    @property
    def url(self) -> str:
        return self.__document.url if self.__document else ""

    def clear(self) -> None:
        """
        Clear the document.
        """
        self.__document = None
        self.__scroll_offset = 0
        self.__invalidate_cache()

    def draw_inner_footer(self, stdscr: curses.window, bounds: ViewBounds, text: str) -> None:
        """
        Draw document footer with scroll position and mode switcher.
        
        Args:
            stdscr: The curses window to draw on
            bounds: The view bounds defining the drawing area
            text: The footer text to display
        """
        if not self.__document:
            super().draw_inner_footer(stdscr, bounds, text)
            return

        style: int = self._get_inner_header_style()
        footer_y: int = bounds.y + bounds.height - 1

        terminal_height: int
        terminal_height, _ = stdscr.getmaxyx()
        if footer_y >= terminal_height:
            return

        content_lines: list[str] = self.__get_content_lines()
        content_height: int = max(0, bounds.height - 4)
        total_lines: int = len(content_lines)
        showing_start: int = self.__scroll_offset + 1
        showing_end: int = min(total_lines, self.__scroll_offset + content_height)
        left_info: str = f"Viewing lines {showing_start}-{showing_end} of {total_lines}"
        modes: list[tuple[str, DocumentMode]] = [
            (" MD ", DocumentMode.MARKDOWN),
            (" RAW ", DocumentMode.RAW),
            (" HDR ", DocumentMode.HEADERS)
        ]

        mode_buttons_width: int = sum(len(mode_name) for mode_name, _ in modes)

        mode_start_x: int = bounds.width - mode_buttons_width - 1
        document_mode_style: int = self.session.get_theme_color_pair(ThemeDefinition.DOCUMENT_MODE)
        safe_addstr(stdscr, footer_y, 0, self._get_bounded_line(), style)
        safe_addstr(stdscr, footer_y, 1, left_info, style)
        if mode_start_x > len(left_info) + 3:
            current_x: int = mode_start_x
            for mode_name, mode_enum in modes:
                is_current: bool = self.__document_mode == mode_enum
                mode_style: int = document_mode_style if is_current else style
                if current_x + len(mode_name) <= bounds.width:
                    safe_addstr(stdscr, footer_y, current_x, mode_name, mode_style)
                current_x += len(mode_name)

    def handle_input(self, key: int) -> bool:
        """
        Handle document navigation input.
        
        Args:
            key: The curses key code from user input
            
        Returns:
            bool: True if the input was handled, False otherwise
        """
        if not self._focused or not self.__document:
            return False

        handlers: dict[int, callable] = {
            curses.KEY_UP: self.__scroll_up,
            curses.KEY_DOWN: self.__scroll_down,
            curses.KEY_LEFT: self.__jump_to_previous_highlight,
            curses.KEY_RIGHT: self.__jump_to_next_highlight,
            curses.KEY_PPAGE: lambda: self.__scroll_page_up(max(1, self.bounds.height - 4)),
            curses.KEY_NPAGE: lambda: self.__scroll_page_down(max(1, self.bounds.height - 4)),
            curses.KEY_HOME: self.__scroll_to_top,
            curses.KEY_END: self.__scroll_to_bottom,
            ord('\t'): self.__cycle_mode,
        }

        handler = handlers.get(key)
        if handler:
            handler()
            return True

        return False

    def render(self, stdscr: curses.window) -> None:
        """
        Render the document view within bounds with search highlighting.
        
        Args:
            stdscr: The curses window to draw on
        """
        if not self._renderable(stdscr):
            return
        if not self.__document:
            self.__render_no_document(stdscr)
            return

        xb: int = self.bounds.x
        yb: int = self.bounds.y
        y_current: int = yb + 2
        y_max: int = yb + self.bounds.height

        content_height: int = max(0, self.bounds.height - 4)
        content_width: int = self.bounds.width - 4
        content_lines: list[str] = self.__get_content_lines()
        visible_lines: list[str] = content_lines[self.__scroll_offset: self.__scroll_offset + content_height]

        self.__update_search_terms()

        for i, line in enumerate(visible_lines):
            line_y: int = y_current + i
            if line_y >= self.bounds.height:
                break

            if self.__search_terms and line.strip():
                self.__render_line_with_highlights(stdscr, line, line_y, 2, content_width)
            else:
                display_line: str = line[:content_width] if len(line) > content_width else line
                safe_addstr(stdscr, line_y, 2, display_line)

    def update(self, document: ResourceResult) -> None:
        """
        Update the document and reset scroll position.
        
        Args:
            document: The resource result document to display
        """
        self.__document = document
        self.__scroll_offset = 0
        self.__invalidate_cache()

    def __calculate_max_scroll(self) -> int:
        """
        Calculate maximum scroll offset based on content and view size.
        
        Returns:
            int: The maximum scroll offset value
        """
        if not self.__document:
            return 0

        content_lines: list[str] = self.__get_content_lines()
        content_height: int = max(0, self.bounds.height - 4)

        return max(0, len(content_lines) - content_height)

    def __cycle_mode(self) -> None:
        """
        Cycle to the next document mode.
        """
        self.__document_mode = DOCUMENT_MODE_NEXT.get(
            self.__document_mode,
            DocumentMode.MARKDOWN
        )
        self.__scroll_offset = 0
        self.__invalidate_cache()

    def __get_content_lines(self) -> list[str]:
        """
        Get content lines based on current mode with caching.
        
        Returns:
            list[str]: The content lines for the current document mode
        """
        current_query: str = self.session.searchform.query if hasattr(self.session, 'searchform') else ""

        if (self.__cached_content_lines is not None and
            self.__cached_mode == self.__document_mode and
            self.__cached_query == current_query):
            return self.__cached_content_lines

        if not self.__document:
            return []

        content_lines: list[str]
        if self.__document_mode == DocumentMode.MARKDOWN:
            content_lines = self.__get_markdown_lines()
        elif self.__document_mode == DocumentMode.RAW:
            content_lines = self.__get_raw_lines()
        elif self.__document_mode == DocumentMode.HEADERS:
            content_lines = self.__get_header_lines()
        else:
            content_lines = ["Unknown document mode"]

        self.__cached_content_lines = content_lines
        self.__cached_mode = self.__document_mode
        self.__cached_query = current_query

        return content_lines

    def __get_header_lines(self) -> list[str]:
        """
        Get headers with proper wrapping.
        
        Returns:
            list[str]: The wrapped header lines
        """
        if not self.__document.headers:
            return ["No headers available for this resource."]

        return self.__wrap_text_content(self.__document.headers)

    def __get_markdown_lines(self) -> list[str]:
        """
        Get markdown content with proper wrapping.
        
        Returns:
            list[str]: The wrapped markdown content lines
        """
        raw_markdown: str = self.__document.get_extra("markdown")
        if not raw_markdown:
            return ["", "Markdown unavailable for this resource."]

        return self.__wrap_text_content(raw_markdown)

    def __get_raw_lines(self) -> list[str]:
        """
        Get raw content with proper wrapping.
        
        Returns:
            list[str]: The wrapped raw content lines
        """
        if not self.__document.content:
            return ["No raw content available for this resource."]

        return self.__wrap_text_content(self.__document.content.strip())

    def __invalidate_cache(self) -> None:
        """
        Invalidate cached content lines.
        """
        self.__cached_content_lines = None
        self.__cached_mode = None
        self.__cached_query = None

    def __jump_to_next_highlight(self) -> None:
        """
        Jump to next highlight, positioning it at line 5 of screen.
        """
        if not self.__search_terms:
            return

        content_lines: list[str] = self.__get_content_lines()
        current_line: int = self.__scroll_offset + 3

        for line_num in range(current_line + 1, len(content_lines)):
            highlights: list[HighlightSpan] = HighlightProcessor.find_highlights_in_text(
                content_lines[line_num],
                self.__search_terms
            )
            if highlights:
                self.__scroll_offset = max(0, line_num - 3)
                return

        for line_num in range(0, current_line + 1):
            highlights: list[HighlightSpan] = HighlightProcessor.find_highlights_in_text(
                content_lines[line_num],
                self.__search_terms
            )
            if highlights:
                self.__scroll_offset = max(0, line_num - 3)
                return

    def __jump_to_previous_highlight(self) -> None:
        """
        Jump to previous highlight, positioning it at line 5 of screen.
        """
        if not self.__search_terms:
            return

        content_lines: list[str] = self.__get_content_lines()
        current_line: int = self.__scroll_offset + 3

        for line_num in range(current_line - 1, -1, -1):
            highlights: list[HighlightSpan] = HighlightProcessor.find_highlights_in_text(
                content_lines[line_num],
                self.__search_terms
            )
            if highlights:
                self.__scroll_offset = max(0, line_num - 3)
                return

        for line_num in range(len(content_lines) - 1, current_line - 1, -1):
            highlights: list[HighlightSpan] = HighlightProcessor.find_highlights_in_text(
                content_lines[line_num],
                self.__search_terms
            )
            if highlights:
                self.__scroll_offset = max(0, line_num - 3)
                return

    def __render_line_with_highlights(self, stdscr: curses.window, line: str, y: int, x: int, max_width: int) -> None:
        """
        Render a line with search term highlighting using the shared utility.
        
        Args:
            stdscr: The curses window to draw on
            line: The text line to render
            y: Y position to render at
            x: X position to render at
            max_width: Maximum width for rendering
        """
        if not line.strip():
            return

        highlights: list[HighlightSpan] = HighlightProcessor.find_highlights_in_text(line, self.__search_terms)
        normal_style: int = curses.A_NORMAL
        highlight_style: int = self.session.get_theme_color_pair(ThemeDefinition.SNIPPET_HIGHLIGHT)
        HighlightProcessor.render_text_with_highlights(
            stdscr, line, highlights, x, y, max_width, normal_style, highlight_style
        )

    def __render_no_document(self, stdscr: curses.window) -> None:
        """
        Render message when no document is loaded.
        
        Args:
            stdscr: The curses window to draw on
        """
        x: int = self.bounds.x
        y: int = self.bounds.y
        width: int = self.bounds.width
        height: int = self.bounds.height

        if height > 2 and width > 20:
            safe_addstr(stdscr, y + 2, x + 2, "No document loaded.", curses.A_DIM)

    def __scroll_down(self, lines: int = 1) -> None:
        """
        Scroll down by specified number of lines.
        
        Args:
            lines: Number of lines to scroll down
        """
        max_scroll: int = self.__calculate_max_scroll()
        self.__scroll_offset = min(max_scroll, self.__scroll_offset + lines)

    def __scroll_page_down(self, page_size: int = 10) -> None:
        """
        Scroll down by page.
        
        Args:
            page_size: Number of lines to scroll for a page
        """
        self.__scroll_down(page_size)

    def __scroll_page_up(self, page_size: int = 10) -> None:
        """
        Scroll up by page.
        
        Args:
            page_size: Number of lines to scroll for a page
        """
        self.__scroll_up(page_size)

    def __scroll_to_bottom(self) -> None:
        """
        Scroll to bottom of document.
        """
        self.__scroll_offset = self.__calculate_max_scroll()

    def __scroll_to_top(self) -> None:
        """
        Scroll to top of document.
        """
        self.__scroll_offset = 0

    def __scroll_up(self, lines: int = 1) -> None:
        """
        Scroll up by specified number of lines.
        
        Args:
            lines: Number of lines to scroll up
        """
        self.__scroll_offset = max(0, self.__scroll_offset - lines)

    def __update_search_terms(self) -> None:
        """
        Update search terms from current search form query using shared utility.
        """
        if hasattr(self.session, 'searchform') and self.session.searchform:
            query: str = self.session.searchform.query
            self.__search_terms = HighlightProcessor.extract_search_terms(query)
        else:
            self.__search_terms = []

    def __wrap_text_content(self, raw_text: str) -> list[str]:
        """
        Wrap text content for display with proper line handling.
        
        Args:
            raw_text: The raw text content to wrap
            
        Returns:
            list[str]: The wrapped text lines
        """
        if not raw_text:
            return []

        content_width: int = max(20, self.bounds.width - CONTENT_MARGIN)
        wrapped_lines: list[str] = []
        text_lines: list[str] = raw_text.split("\n")

        for line in text_lines:
            if not line.strip():
                wrapped_lines.append("")
            else:
                wrapped: str = textwrap.fill(
                    line.rstrip(),
                    width=content_width,
                    expand_tabs=True,
                    replace_whitespace=False,
                    break_long_words=True,
                    break_on_hyphens=True
                )
                wrapped_lines.extend(wrapped.split("\n"))

        return wrapped_lines



from mcp_server_webcrawl.crawlers import get_fixture_directory
from mcp_server_webcrawl.crawlers.wget.adapter import WgetManager
from mcp_server_webcrawl.crawlers.wget.crawler import WgetCrawler
from mcp_server_webcrawl.crawlers.base.tests import BaseCrawlerTests
from mcp_server_webcrawl.utils.logger import get_logger

logger = get_logger()

EXAMPLE_SITE_ID = WgetManager.string_to_id("example.com")
PRAGMAR_SITE_ID = WgetManager.string_to_id("pragmar.com")

class WgetTests(BaseCrawlerTests):
    """
    Test suite for the wget crawler implementation.
    Uses all wrapped test methods from BaseCrawlerTests.
    """

    def setUp(self):
        """
        Set up the test environment with fixture data.
        """
        super().setUp()
        self._datasrc = get_fixture_directory() / "wget"

    def test_wget_pulse(self):
        """
        Test basic crawler initialization.
        """
        crawler = WgetCrawler(self._datasrc)
        self.assertIsNotNone(crawler)
        self.assertTrue(self._datasrc.is_dir())

    def test_wget_sites(self):
        """
        Test site retrieval API functionality.
        """
        crawler = WgetCrawler(self._datasrc)
        self.run_pragmar_site_tests(crawler, PRAGMAR_SITE_ID)

    def test_wget_search(self):
        """
        Test boolean search functionality
        """
        # moved fixtures to own repo, lost some local media,
        # but checks out. wget fixture has no CSS/JS/etc.
        # HTML-only and just doesn't do well with the full array of
        # tests concerning fulltext, media, and mixed search result
        # counts. probably needs a reduced set of tests
        # self.run_pragmar_search_tests(crawler, PRAGMAR_SITE_ID)
        return

    def test_wget_resources(self):
        """
        Test resource retrieval API functionality with various parameters.
        """
        crawler = WgetCrawler(self._datasrc)
        self.run_sites_resources_tests(crawler, PRAGMAR_SITE_ID, EXAMPLE_SITE_ID)


    def test_wget_sorts(self):
        """
        Test random sort functionality using the '?' sort parameter.
        """
        crawler = WgetCrawler(self._datasrc)
        self.run_pragmar_sort_tests(crawler, PRAGMAR_SITE_ID)

    def test_wget_content_parsing(self):
        """
        Test content type detection and parsing.
        """
        crawler = WgetCrawler(self._datasrc)
        self.run_pragmar_content_tests(crawler, PRAGMAR_SITE_ID, False)

    def test_report(self):
        """
        Run test report, save to data directory.
        """
        crawler = WgetCrawler(self._datasrc)
        logger.info(self.run_pragmar_report(crawler, PRAGMAR_SITE_ID, "wget"))



import re
import curses

from dataclasses import dataclass
from typing import List

from mcp_server_webcrawl.interactive.ui import safe_addstr

@dataclass
class HighlightSpan:
    """
    Represents a highlight span in text
    """
    start: int
    end: int
    text: str

    def __str__(self) -> str:
        return f"[{self.start}:{self.end} '{self.text}']"


class HighlightProcessor:
    """
    Shared highlight processing utilities
    """

    QUOTED_PHRASE_PATTERN = re.compile(r'"([^"]+)"')
    WORD_PATTERN = re.compile(r"\b\w+\b")
    SNIPPET_MARKER_PATTERN = re.compile(r"\*\*([a-zA-Z\-_' ]+)\*\*")
    IGNORE_WORDS = {"AND", "OR", "NOT", "and", "or", "not", "type", "status", "size", "url", "id"}

    @staticmethod
    def extract_search_terms(query: str) -> List[str]:
        """
        Extract search terms from query, handling quoted phrases and individual keywords.
        """
        if not query or not query.strip():
            return []

        search_terms = []
        for match in HighlightProcessor.QUOTED_PHRASE_PATTERN.finditer(query):
            phrase = match.group(1).strip()
            if phrase:
                search_terms.append(phrase)

        remaining_query = HighlightProcessor.QUOTED_PHRASE_PATTERN.sub('', query)

        # extract individual words
        for match in HighlightProcessor.WORD_PATTERN.finditer(remaining_query):
            word = match.group().strip()
            if word and word not in HighlightProcessor.IGNORE_WORDS and len(word) > 2:
                search_terms.append(word)

        return search_terms

    @staticmethod
    def find_highlights_in_text(text: str, search_terms: List[str]) -> List[HighlightSpan]:
        """
        Find all highlight spans in text for the given search terms.
        """
        if not text or not search_terms:
            return []

        highlights = []
        escaped_terms = [re.escape(term.strip("\"'")) for term in search_terms]
        pattern = re.compile(rf"\b({'|'.join(escaped_terms)})\b", re.IGNORECASE)

        for match in pattern.finditer(text):
            span = HighlightSpan(
                start=match.start(),
                end=match.end(),
                text=match.group()
            )
            highlights.append(span)

        return HighlightProcessor.merge_overlapping_highlights(highlights, text)

    @staticmethod
    def extract_snippet_highlights(snippet_text: str) -> tuple[str, List[HighlightSpan]]:
        """
        Extract highlights from snippet text with **markers**, returning clean text and highlights.
        """
        if not snippet_text:
            return "", []

        normalized_text = re.sub(r"\s+", " ", snippet_text.strip())

        clean_text = ""
        highlights = []
        last_end = 0

        for match in HighlightProcessor.SNIPPET_MARKER_PATTERN.finditer(normalized_text):
            # text before this match
            clean_text += normalized_text[last_end:match.start()]

            # highlighted text (without markers)
            highlight_text = match.group(1)
            highlight_start = len(clean_text)
            clean_text += highlight_text
            highlight_end = len(clean_text)

            span: HighlightSpan = HighlightSpan(
                start=highlight_start,
                end=highlight_end,
                text=highlight_text
            )
            highlights.append(span)
            last_end = match.end()

        # remaining text
        clean_text += normalized_text[last_end:]

        return clean_text.strip(), highlights

    @staticmethod
    def merge_overlapping_highlights(highlights: List[HighlightSpan], text: str) -> List[HighlightSpan]:
        """Merge overlapping or adjacent highlight spans."""
        if not highlights:
            return []

        # sort by start position
        sorted_highlights = sorted(highlights, key=lambda h: h.start)
        merged = []

        for highlight in sorted_highlights:
            if not merged:
                merged.append(highlight)
            else:
                last = merged[-1]
                if highlight.start <= last.end:
                    # overlapping/adjacent - merge them
                    end = max(last.end, highlight.end)
                    merged_text = text[last.start:end]
                    merged[-1] = HighlightSpan(
                        start=last.start,
                        end=end,
                        text=merged_text
                    )
                else:
                    merged.append(highlight)

        return merged

    @staticmethod
    def render_text_with_highlights(
        stdscr: curses.window,
        text: str,
        highlights: List[HighlightSpan],
        x: int,
        y: int,
        max_width: int,
        normal_style: int,
        hit_style: int
    ) -> None:
        """
        Render text with highlights applied.
        """
        if not text.strip():
            return

        display_text: str = text[:max_width] if len(text) > max_width else text
        visible_highlights: list[str] = [h for h in highlights if h.start < len(display_text)]
        current_x: int = x
        pos: int = 0

        try:
            for highlight in visible_highlights:
                # text before highlight
                if highlight.start > pos:
                    text_before: str = display_text[pos:highlight.start]
                    safe_addstr(stdscr, y, current_x, text_before, normal_style)
                    current_x += len(text_before)
                    pos = highlight.start

                # highlighted text
                highlight_end: int = min(highlight.end, len(display_text))
                highlighted_text: str = display_text[highlight.start:highlight_end]
                if current_x + len(highlighted_text) <= x + max_width:
                    safe_addstr(stdscr, y, current_x, highlighted_text, hit_style)
                    current_x += len(highlighted_text)
                pos = highlight_end

            # remaining text
            if pos < len(display_text):
                remaining_text: str = display_text[pos:]
                remaining_width: int = max_width - (current_x - x)
                if remaining_width > 0:
                    safe_addstr(stdscr, y, current_x, remaining_text[:remaining_width], normal_style)

        except curses.error:
            pass



import os
import sys
import asyncio
import tracemalloc
import unittest
import logging

from pathlib import Path
from argparse import ArgumentParser

from mcp_server_webcrawl.utils.cli import get_help_short_message, get_help_long_message
from mcp_server_webcrawl.settings import DEBUG, DATA_DIRECTORY, FIXTURES_DIRECTORY
from mcp_server_webcrawl.crawlers import get_crawler, VALID_CRAWLER_CHOICES

__version__: str = "0.15.1"
__name__: str = "mcp-server-webcrawl"

if DEBUG:
    tracemalloc.start()

class CustomHelpArgumentParser(ArgumentParser):
    def print_help(self, file=None):
        print(get_help_long_message(__version__))

def main() -> None:
    """
    Main entry point for the package. mcp-server-webcrawl should be on path if pip installed
    """

    if len(sys.argv) == 1:
        # \n parser error follows short message
        sys.stderr.write(get_help_short_message(__version__) + "\n")

    parser: CustomHelpArgumentParser = CustomHelpArgumentParser(description="InterrBot MCP Server")
    parser.add_argument("-c", "--crawler", type=str, choices=VALID_CRAWLER_CHOICES,
            help="Specify which crawler to use (default: interrobot)")
    parser.add_argument("--run-tests", action="store_true", help="Run tests instead of server")
    parser.add_argument("-i", "--interactive", action="store_true", help="Run interactive terminal search mode")
    parser.add_argument("-d", "--datasrc", type=str, help="Path to datasrc (required unless testing)")
    args = parser.parse_args()

    if args.run_tests:

        # Check if FIXTURES_DIRECTORY is configured and exists
        if FIXTURES_DIRECTORY is None or not FIXTURES_DIRECTORY.exists() or not FIXTURES_DIRECTORY.is_dir():
            sys.stderr.write(f"Fixtures not configured in settings_local.py, or is not a valid directory.\nFIXTURES_DIRECTORY: {FIXTURES_DIRECTORY}")
            sys.exit(1)

        # testing captures some cross-fixture file information, useful for debug
        # force=True gets this to write during tests (usually quieted during run)
        unittest_log: Path = DATA_DIRECTORY / "fixtures-report.log"
        logging.basicConfig(level=logging.INFO, filename=unittest_log, filemode='w', force=True)
        file_directory = os.path.dirname(os.path.abspath(__file__))
        sys.exit(unittest.main(module=None, argv=["", "discover", "-s", file_directory, "-p", "*test*.py"]))

    if args.interactive:
        from mcp_server_webcrawl.interactive.session import InteractiveSession
        intersession = InteractiveSession(args.crawler, args.datasrc)
        intersession.run()
        sys.exit(0)

    if not args.datasrc:
        parser.error("the -d/--datasrc argument is required when not in test mode")

    if not args.crawler or args.crawler.lower() not in VALID_CRAWLER_CHOICES:
        valid_crawlers = ", ".join(VALID_CRAWLER_CHOICES)
        parser.error(f"the -c/--crawler argument must be one of: {valid_crawlers}")

    # cli interaction prior to loading the server
    from mcp_server_webcrawl.main import main as mcp_main
    crawler = get_crawler(args.crawler)
    asyncio.run(mcp_main(crawler, Path(args.datasrc)))

__all__ = ["main"]



from pathlib import Path

from mcp_server_webcrawl.crawlers.base.indexed import IndexedCrawler
from mcp_server_webcrawl.crawlers.katana.adapter import get_sites, get_resources
from mcp_server_webcrawl.utils.logger import get_logger

logger = get_logger()

class KatanaCrawler(IndexedCrawler):
    """
    A crawler implementation for HTTP text files.
    Provides functionality for accessing and searching web content from captured HTTP exchanges.
    """

    def __init__(self, datasrc: Path):
        """
        Initialize the HTTP text crawler with a data source directory.

        Args:
            datasrc: The input argument as Path, it must be a directory containing
                subdirectories with HTTP text files
        """
        super().__init__(datasrc, get_sites, get_resources)



import re
import curses

from abc import abstractmethod
from typing import TYPE_CHECKING

from mcp_server_webcrawl import __name__ as module_name, __version__ as module_version
from mcp_server_webcrawl.interactive.ui import ThemeDefinition, ViewBounds
from mcp_server_webcrawl.models.resources import ResourceResult
from mcp_server_webcrawl.interactive.ui import safe_addstr

if TYPE_CHECKING:
    from mcp_server_webcrawl.interactive.session import InteractiveSession

REGEX_DISPLAY_URL_CLEAN = re.compile(r"^https?://|/$")
OUTER_WIDTH_RIGHT_MARGIN = 1

LAYOUT_FOOTER_SEPARATOR = " | "
LAYOUT_FOOTER_SEPARATOR_LENGTH = len(LAYOUT_FOOTER_SEPARATOR)
MIN_TERMINAL_HEIGHT = 8
MIN_TERMINAL_WIDTH = 40
CONTENT_MARGIN = 4

class BaseCursesView:
    """
    Base class for all views with common interface.
    """

    def __init__(self, session: 'InteractiveSession'):
        self.session = session
        self.bounds = ViewBounds(x=0, y=0, width=0, height=0)
        self._focused = False
        self._selected_index: int = 0

    @property
    def focused(self) -> bool:
        return self._focused

    def set_bounds(self, bounds: ViewBounds):
        """
        Set the rendering bounds for this view.
        
        Args:
            bounds: The ViewBounds object defining the drawing area
        """
        self.bounds = bounds

    def set_focused(self, focused: bool):
        """
        Set the focus state for this view.
        
        Args:
            focused: True if this view should be focused, False otherwise
        """
        self._focused = focused

    @abstractmethod
    def render(self, stdscr: curses.window) -> None:
        """
        Render the view within its bounds.
        
        Args:
            stdscr: The curses window to render on
        """
        pass

    @abstractmethod
    def handle_input(self, key: int) -> bool:
        """
        Handle input. Return True if consumed, False to pass through.
        
        Args:
            key: The input key code
            
        Returns:
            bool: True if input was consumed, False to pass through
        """
        pass

    def focusable(self) -> bool:
        """
        Return True if this view can receive focus.
        
        Returns:
            bool: True if this view can receive focus
        """
        return True

    def draw_outer_footer(self, stdscr: curses.window, text: str) -> None:
        """
        Draw context-sensitive help footer with pipe-separated items.
        
        Args:
            stdscr: The curses window to draw on
            text: The footer text to display (pipe-separated items)
        """
        height, width = stdscr.getmaxyx()
        footer_line: int = height - 1
        footer_line_text: str = BaseCursesView._get_full_width_line(stdscr)
        outer_theme_pair: int = self.session.get_theme_color_pair(ThemeDefinition.HEADER_OUTER)

        safe_addstr(stdscr, footer_line, 0, footer_line_text, outer_theme_pair)
        items = [item.strip() for item in text.split(LAYOUT_FOOTER_SEPARATOR)]
        available_width = width - 4 - 2  # 4 for right margin, 2 for left padding

        display_text: str = ""
        test_text: str = ""
        test_text_length: int = 0
        for i in range(len(items)):
            test_text = LAYOUT_FOOTER_SEPARATOR.join(items[:i+1])
            test_text_length = len(test_text)
            if test_text_length <= available_width:
                display_text = test_text
            else:
                break

         # doesn't fit indicator
        display_text_length: int = len(display_text)
        if test_text_length > available_width:
            display_text += f"{(width - display_text_length - 5) * ' '} »"

        if display_text:
            outer_header_theme_pair: int = self.session.get_theme_color_pair(ThemeDefinition.HEADER_OUTER)
            safe_addstr(stdscr, footer_line, 1, display_text, outer_header_theme_pair)

    def draw_outer_header(self, stdscr: curses.window) -> None:
        """
        Draw the inner header for this view section.
        
        Args:
            stdscr: The curses window to draw on
        """
        _, width = stdscr.getmaxyx()
        style: int = self.session.get_theme_color_pair(ThemeDefinition.HEADER_OUTER)

        full_width_line: str = BaseCursesView._get_full_width_line(stdscr)
        header_label_text: str = f"{module_name} --interactive"
        header_version_text: str = f"v{module_version}"
        header_version_x: int = max(0, width - len(header_version_text) - 2)

        safe_addstr(stdscr, 0, 0, full_width_line, style)
        if len(header_label_text) < width - 2:
            safe_addstr(stdscr, 0, 1, header_label_text, style)

        if header_version_x > len(header_label_text) + 3:
            safe_addstr(stdscr, 0, header_version_x, header_version_text, style)

    def draw_inner_footer(self, stdscr: curses.window, bounds: ViewBounds, text: str) -> None:
        """
        Draw context-sensitive help footer.
        
        Args:
            stdscr: The curses window to draw on
            bounds: The view bounds defining the drawing area
            text: The footer text to display
        """
        footer_y: int = bounds.y + bounds.height - 1
        line_of_whitespace: str = self._get_bounded_line()
        display_text: str = text or ""
        display_text_max: int = len(line_of_whitespace) - 2
        if len(display_text) > display_text_max:
            display_text = f"{display_text[:display_text_max - 1]}…"

        line: str = f" {display_text}".ljust(len(line_of_whitespace))
        safe_addstr(stdscr, footer_y, bounds.x, line, self._get_inner_header_style())

    def draw_inner_header(self, stdscr: curses.window, bounds: ViewBounds, text: str) -> None:
        """
        Draw the application header with module name and version.
        
        Args:
            stdscr: The curses window to draw on
            bounds: The view bounds defining the drawing area
            text: The header text to display
        """

        line_of_whitespace: str = self._get_bounded_line()
        display_text: str = text or ""
        max_text_width: int = len(line_of_whitespace) - 2
        if len(display_text) > max_text_width:
            display_text = f"{display_text[:max_text_width - 1]}…"

        line: str = f" {display_text}".ljust(len(line_of_whitespace))
        safe_addstr(stdscr, bounds.y, bounds.x, line, self._get_inner_header_style())


    @staticmethod
    def _get_full_width_line(stdscr: curses.window) -> str:
        """
        Get a line that fills the terminal width.
        
        Args:
            stdscr: The curses window to get dimensions from
            
        Returns:
            str: A string of spaces filling the terminal width
        """
        _, width = stdscr.getmaxyx()
        return " " * (width - OUTER_WIDTH_RIGHT_MARGIN)

    @staticmethod
    def url_for_display(url: str) -> str:
        """
        Remove protocol prefix and trailing slash from URL for display.
        
        Args:
            url: The URL to clean for display
            
        Returns:
            str: The cleaned URL without protocol and trailing slash
        """
        return REGEX_DISPLAY_URL_CLEAN.sub("", url)

    @staticmethod
    def humanized_bytes(result: ResourceResult) -> str:
        """
        Convert resource size to human-readable format (B, KB, MB).
        
        Args:
            result: The ResourceResult containing size information
            
        Returns:
            str: Human-readable size string (e.g., "1.5MB", "512KB", "128B")
        """
        display: str = ""
        if result is not None:
            size: int = result.size
            if isinstance(size, int):
                if size >= 1024 * 1024:
                    display = f"{size/(1024*1024):.1f}MB"
                elif size >= 1024:
                    display = f"{size/1024:.1f}KB"
                else:
                    display = f"{size}B"
        return display

    def _get_inner_header_style(self) -> int:
        """
        Get the appropriate header style based on focus state.
        
        Returns:
            int: The theme color pair for the header
        """
        if self._focused == True:
            return self.session.get_theme_color_pair(ThemeDefinition.HEADER_ACTIVE)
        else:
            return self.session.get_theme_color_pair(ThemeDefinition.HEADER_INACTIVE)

    def _get_input_style(self) -> int:
        """
        Get the appropriate input style based on focus and selection state.
        
        Returns:
            int: The style attributes for input rendering
        """
        if self._focused and self._selected_index == 0:
            return curses.A_REVERSE
        else:
            return self.session.get_theme_color_pair(ThemeDefinition.INACTIVE_QUERY)

    def _get_bounded_line(self) -> str:
        """
        Get a line of spaces that fits within the view bounds.
        
        Returns:
            str: A string of spaces matching the view width
        """
        return " " * self.bounds.width

    def _renderable(self, stdscr: curses.window) -> bool:
        """
        Check if the view can be rendered within the current terminal bounds.
        
        Args:
            stdscr: The curses window to check dimensions against
            
        Returns:
            bool: True if the view can be rendered, False otherwise
        """
        terminal_height, terminal_width = stdscr.getmaxyx()
        if self.bounds.y >= terminal_height or self.bounds.x >= terminal_width or self.bounds.width <= 0 or self.bounds.height <= 0:
            return False
        return True



from datetime import datetime
from typing import Final
from pathlib import Path
from enum import Enum

from mcp_server_webcrawl.models.base import BaseModel, METADATA_VALUE_TYPE
from mcp_server_webcrawl.utils import to_isoformat_zulu

class SiteType(Enum):
    UNDEFINED = "undefined"
    CRAWLED_URL = "url"
    CRAWLED_LIST = "list"

SITES_TOOL_NAME: Final[str] = "webcrawl_sites"
SITES_FIELDS_BASE: Final[list[str]] = ["id", "name", "type", "urls"]
SITES_FIELDS_DEFAULT: Final[list[str]] = SITES_FIELDS_BASE + ["created", "modified"]

class SiteResult(BaseModel):
    """
    Represents a website or crawl directory result.
    """

    def __init__(
        self,
        id: int,
        name: str | None = None,
        type: SiteType = SiteType.CRAWLED_URL,
        urls: list[str] | None = None,
        path: Path = None,
        created: datetime | None = None,
        modified: datetime | None = None,
        robots: str | None = None,
        metadata: dict[str, METADATA_VALUE_TYPE] | None = None
    ):
        """
        Initialize a SiteResult instance.

        Args:
            id: site identifier
            name: site name, either a URL or a custom job
            urls: site URL(s), multiple for list type crawls
            path: path to site data, different from datasrc
            created: creation timestamp
            modified: last modification timestamp
            robots: robots.txt content
            metadata: additional metadata for the site
        """
        self.id = id
        self.name = name
        self.type = type
        self.urls = urls
        self.path = path
        self.created = created
        self.modified = modified
        self.robots = robots
        self.metadata = metadata or {}

    def to_dict(self) -> dict[str, METADATA_VALUE_TYPE]:
        """
        Convert the object to a dictionary suitable for JSON serialization.
        """
        result: dict[str, METADATA_VALUE_TYPE] = {
            "id": self.id,
            "name": self.name,
            "type": self.type.value,
            "urls": self.urls,
            "created": to_isoformat_zulu(self.created) if self.created else None,
            "modified": to_isoformat_zulu(self.modified) if self.modified else None,
            "metadata": self.metadata if self.metadata else None,
        }

        return {k: v for k, v in result.items() if v is not None and not (k == "metadata" and v == {})}




import sys
from pathlib import Path
from mcp_server_webcrawl.settings import FIXTURES_DIRECTORY

VALID_CRAWLER_CHOICES: list[str] = ["archivebox", "httrack", "interrobot", "katana", "siteone", "warc", "wget"]

def get_fixture_directory() -> Path:
    # only to be used for devs on test runs, configured in settings_local.py
    # settings_local.py added as sibling of settings.py if not present
    # download https://github.com/pragmar/mcp-server-webcrawl-fixtures
    assert FIXTURES_DIRECTORY is not None and FIXTURES_DIRECTORY.is_dir(), \
        f"Fixtures not configured in settings_local.py, or is not a valid directory.\nFIXTURES_DIRECTORY: {FIXTURES_DIRECTORY}"

    return FIXTURES_DIRECTORY

def get_crawler(crawler_name: str) -> str | None:
    """
    lazy load crawler, some classes have additional package dependencies
    """
    if crawler_name is None:
        return None
    crawler_name = crawler_name.lower()
    if crawler_name == "archivebox":
        from mcp_server_webcrawl.crawlers.archivebox.crawler import ArchiveBoxCrawler
        return ArchiveBoxCrawler
    elif crawler_name == "httrack":
        from mcp_server_webcrawl.crawlers.httrack.crawler import HtTrackCrawler
        return HtTrackCrawler
    elif crawler_name == "interrobot":
        from mcp_server_webcrawl.crawlers.interrobot.crawler import InterroBotCrawler
        return InterroBotCrawler
    elif crawler_name == "katana":
        from mcp_server_webcrawl.crawlers.katana.crawler import KatanaCrawler
        return KatanaCrawler
    elif crawler_name == "siteone":
        from mcp_server_webcrawl.crawlers.siteone.crawler import SiteOneCrawler
        return SiteOneCrawler
    elif crawler_name == "warc":
        from mcp_server_webcrawl.crawlers.warc.crawler import WarcCrawler
        return WarcCrawler
    elif crawler_name == "wget":
        from mcp_server_webcrawl.crawlers.wget.crawler import WgetCrawler
        return WgetCrawler
    else:
        valid_choices = ", ".join(VALID_CRAWLER_CHOICES)
        raise ValueError(f"unsupported crawler '{crawler_name}' ({valid_choices})")



import hashlib
import threading
from concurrent.futures import ThreadPoolExecutor, Future
from datetime import datetime
from typing import Optional, TYPE_CHECKING

from mcp_server_webcrawl.crawlers.base.crawler import BaseJsonApi
from mcp_server_webcrawl.interactive.ui import UiFocusable, UiState
from mcp_server_webcrawl.models.resources import ResourceResult

if TYPE_CHECKING:
    from mcp_server_webcrawl.interactive.session import InteractiveSession

SEARCH_DEBOUNCE_DELAY_SECONDS = 0.2
SEARCH_RESULT_LIMIT: int = 10

class SearchManager:
    """
    Manages search operations including async search and debouncing.
    Works with session's controlled interface - never touches private state directly.
    """

    def __init__(self, session: 'InteractiveSession'):
        self.__session: 'InteractiveSession' = session
        self.__search_last_state_hash: str = ""
        self.__search_timer: Optional[threading.Timer] = None
        self.__executor = ThreadPoolExecutor(max_workers=2, thread_name_prefix="SearchManager")
        self.__search_lock: threading.RLock = threading.RLock()
        self.__search_in_progress: bool = False
        self.__active_search_future: Optional[Future] = None
        self.__pending_results: Optional[list[ResourceResult]] = None
        self.__pending_indexer_status: str = ""
        self.__pending_indexer_processed: int = 0
        self.__pending_indexer_duration: float = 0
        self.__pending_total: int = 0

    def autosearch(self, immediate: bool = False) -> None:
        """
        Trigger search with optional immediate execution.
        
        Args:
            immediate: If True, execute search synchronously without debouncing.
                    If False, use debounced async execution (default).
        """
        current_state_hash: str = self.__get_input_hash()

        if not immediate and current_state_hash == self.__search_last_state_hash:
            return

        self.__search_last_state_hash = current_state_hash
        self.cancel_pending()

        if immediate:
            self.__execute_search_immediate()
        else:
            self.__search_timer = threading.Timer(SEARCH_DEBOUNCE_DELAY_SECONDS, self.__execute_debounced_search)
            self.__search_timer.start()

    def cancel_pending(self) -> None:
        """
        Cancel any pending search timer.
        """
        if self.__search_timer is not None:
            self.__search_timer.cancel()
            self.__search_timer = None

        with self.__search_lock:
            if self.__active_search_future is not None:
                self.__active_search_future.cancel()
                self.__active_search_future = None

    def check_pending(self) -> None:
        """
        Check if there are pending search results and update the UI.
        """
        with self.__search_lock:
            if self.__pending_results is not None:
                self.__session.results.update(self.__pending_results, self.__pending_total, self.__pending_indexer_status,
                        self.__pending_indexer_processed, self.__pending_indexer_duration)
                self.__pending_results = None
                self.__pending_total = 0
                self.__pending_indexer_processed = 0
                self.__pending_indexer_duration = 0

    def cleanup(self) -> None:
        """
        Clean up any pending operations.
        """
        self.cancel_pending()
        self.__executor.shutdown(wait=True)

    def has_pending(self) -> bool:
        """
        Check if there's a pending debounced search.
        """
        return self.__search_timer is not None

    def is_searching(self) -> bool:
        """
        Check if a search is currently in progress or on a timer.
        """
        with self.__search_lock:
            return self.__search_in_progress or self.__search_timer is not None

    def __background_search(self) -> None:
        """
        Execute search in background thread and store results.
        """
        with self.__search_lock:
            self.__search_in_progress = True

        self.__session.searchform.set_search_attempted()
        results, total_results, index_status, index_processed_count, index_duration_value = self.__execute_search_query()
        self.__set_pending_results(results, total_results, index_status, index_processed_count, index_duration_value, False)

    def __build_search_query(self, base_query: str) -> str:
        """
        Build the final search query with filter applied (if present).        
        """
        if self.__session.searchform.filter == "html":
            if base_query.strip():
                return f"(type: html) AND {base_query}"
            else:
                return "type: html"
        else:
            return base_query

    def __execute_debounced_search(self) -> None:
        """
        Execute search after debounce delay in separate thread.
        """
        current_state_hash: str = self.__get_input_hash()
        if current_state_hash != self.__search_last_state_hash:
            return

        # show split view on results
        if self.__session.ui_focused == UiFocusable.SEARCH_RESULTS:
            self.__session.set_ui_state(UiState.SEARCH_RESULTS, UiFocusable.SEARCH_RESULTS)
        else:
            self.__session.set_ui_state(UiState.SEARCH_RESULTS, UiFocusable.SEARCH_FORM)

        self.__search_timer = None
        with self.__search_lock:
            self.__active_search_future = self.__executor.submit(self.__background_search)

    def __execute_search_immediate(self) -> None:
        """
        Execute search immediately on main thread (for ENTER key).
        """
        self.__session.searchform.set_search_attempted()

        self.__set_pending_results(None, 0, "", -1, -1, False)
        self.__session.results.clear()
        results, total_results, index_status, index_processed_count, index_duration_value = self.__execute_search_query()
        self.__set_pending_results(results, total_results, index_status, index_processed_count, index_duration_value, False)

    def __execute_search_query(self) -> tuple[list[ResourceResult], int, str, int, float]:
        """
        Centralized search execution logic shared by both sync and async paths.
        
        Returns:
            tuple: (results, total_results, index_status, index_processed_count, index_duration_value)
        """
        api: BaseJsonApi | None = self.__get_results(offset=self.__session.searchform.offset)

        if api is None:
            return [], 0, 0, 0

        results: list[ResourceResult] = api.get_results()
        total_results: int = api.total

        index_status: str = ""
        index_processed_count: int = -1
        index_duration_value: float = -1

        if api.meta_index is not None:
            if "status" in api.meta_index:
                index_status = api.meta_index["status"]
            if "processed" in api.meta_index:
                index_processed_count = api.meta_index["processed"]

            if "duration" in api.meta_index:
                index_duration_string: str = api.meta_index["duration"] or ""
                if index_duration_string:
                    try:
                        dt: datetime = datetime.strptime(index_duration_string, "%H:%M:%S.%f")
                        index_duration_value = dt.hour * 3600 + dt.minute * 60 + dt.second + dt.microsecond / 1000000
                    except ValueError:
                        index_duration_value = 0

        return results, total_results, index_status, index_processed_count, index_duration_value

    def __get_input_hash(self) -> str:
        """
        Generate a hash representing the complete current search state.
        """
        query: str = self.__session.searchform.query.strip()
        selected_sites = self.__session.searchform.get_selected_sites()
        selected_sites_ids: list[int] = [s.id for s in selected_sites]
        filter: str = str(self.__session.searchform.filter)
        sort: str = str(self.__session.searchform.sort)
        offset: int = self.__session.searchform.offset
        limit: int = self.__session.searchform.limit
        search_state: str = f"{query}|{selected_sites_ids}|{filter}|{offset}|{limit}|{sort}"
        return hashlib.md5(search_state.encode()).hexdigest()

    def __get_results(self, offset: int = 0) -> BaseJsonApi | None:
        """
        Execute search with given offset and return API response object.
        Centralizes the API call logic used by both sync and async search paths.
        
        Args:
            offset: Starting position for search results pagination
            
        Returns:
            BaseJsonApi: API response object containing search results and metadata
        """
        selected_site_ids: list[int] = self.__get_selected_site_ids()

        query: str = self.__build_search_query(self.__session.searchform.query)
        sort: str = self.__session.searchform.sort
        query_api: BaseJsonApi = self.__session.crawler.get_resources_api(
            sites=selected_site_ids if selected_site_ids else None,
            query=query,
            fields=["size", "status"],
            offset=offset,
            limit=SEARCH_RESULT_LIMIT,
            extras=["snippets"],
            sort=sort
        )

        return query_api

    def __get_selected_site_ids(self) -> list[int]:
        """
        Get list of selected site IDs using property access.
        """
        selected_sites = self.__session.searchform.get_selected_sites()
        return [site.id for site in selected_sites]

    def __set_pending_results(self, results, total_results, index_status, index_processed_count, index_duration_value, search_in_progress) -> None:
        try:
            with self.__search_lock:
                self.__pending_results = results
                self.__pending_total = total_results
                self.__pending_indexer_status = index_status
                self.__pending_indexer_processed = index_processed_count
                self.__pending_indexer_duration = index_duration_value
                self.__search_in_progress = search_in_progress

        except Exception as ex:
            with self.__search_lock:
                self.__session.results.clear()
                self.__search_in_progress = False



import os
import re
import sqlite3

from contextlib import closing
from datetime import datetime, timezone
from pathlib import Path

from mcp_server_webcrawl.crawlers.base.adapter import (
    BaseManager,
    IndexState,
    IndexStatus,
    SitesGroup,
    INDEXED_BATCH_SIZE,
    INDEXED_BYTE_MULTIPLIER,
    INDEXED_RESOURCE_DEFAULT_PROTOCOL,
    INDEXED_TYPE_MAPPING,
)
from mcp_server_webcrawl.crawlers.base.indexed import IndexedManager
from mcp_server_webcrawl.utils.logger import get_logger
from mcp_server_webcrawl.models.resources import (
    ResourceResult,
    ResourceResultType,
    RESOURCES_LIMIT_DEFAULT,
)
from mcp_server_webcrawl.models.sites import (
    SiteResult,
)

SITEONE_LOG_TYPE_MAPPING = {
    "html": ResourceResultType.PAGE,
    "redirect": ResourceResultType.PAGE,
    "image": ResourceResultType.IMAGE,
    "js": ResourceResultType.SCRIPT,
    "css": ResourceResultType.CSS,
    "video": ResourceResultType.VIDEO,
    "audio": ResourceResultType.AUDIO,
    "pdf": ResourceResultType.PDF,
    "other": ResourceResultType.OTHER,
    "font": ResourceResultType.OTHER,
}

logger = get_logger()

class SiteOneManager(IndexedManager):
    """
    Manages SiteOne directory data in in-memory SQLite databases.
    Wraps wget archive format (shared by SiteOne and wget)
    Provides connection pooling and caching for efficient access.
    """

    def __init__(self) -> None:
        """Initialize the SiteOne manager with empty cache and statistics."""

        super().__init__()

    def _extract_log_metadata(self, directory: Path) -> tuple[dict, dict]:
        """
        Extract metadata from SiteOne log files.

        Args:
            directory: path to the site directory

        Returns:
            Tuple of (success log data, error log data) dictionaries
        """
        directory_name: str = directory.name
        log_data = {}
        log_http_error_data = {}

        log_pattern: str = f"output.{directory_name}.*.txt"
        log_files = list(Path(directory.parent).glob(log_pattern))

        if not log_files:
            return log_data, log_http_error_data

        log_latest = max(log_files, key=lambda p: p.stat().st_mtime)

        try:
            with open(log_latest, "r", encoding="utf-8") as log_file:
                for line in log_file:
                    parts = [part.strip() for part in line.split("|")]
                    if len(parts) == 10:
                        parts_path = parts[3].split("?")[0]
                        try:
                            status = int(parts[4])
                            url = f"{INDEXED_RESOURCE_DEFAULT_PROTOCOL}{directory_name}{parts_path}"
                            time_str = parts[6].split()[0]
                            time = int(float(time_str) * (1000 if "s" in parts[6] else 1))

                            # size collected for errors, os stat preferred
                            size_str = parts[7].strip()
                            size = 0
                            if size_str:
                                size_value = float(size_str.split()[0])
                                size_unit = size_str.split()[1].lower() if len(size_str.split()) > 1 else "b"
                                multiplier = INDEXED_BYTE_MULTIPLIER.get(size_unit, 1)
                                size = int(size_value * multiplier)

                            if 400 <= status < 600:
                                log_http_error_data[url] = {
                                    "status": status,
                                    "type": parts[5].lower(),
                                    "time": time,
                                    "size": size,
                                }
                            else:
                                log_data[url] = {
                                    "status": status,
                                    "type": parts[5].lower(),
                                    "time": time,
                                    "size": size,
                                }

                        except (ValueError, IndexError, UnicodeDecodeError, KeyError):
                            continue

                    elif line.strip() == "Redirected URLs":
                        # stop processing we're through HTTP requests
                        break
        except Exception as ex:
            logger.error(f"Error processing log file {log_latest}: {ex}")

        return log_data, log_http_error_data

    def _load_site_data(self, connection: sqlite3.Connection, directory: Path,
            site_id: int, index_state: IndexState = None) -> None:
        """
        Load a SiteOne directory into the database with parallel processing and batch insertions.

        Args:
            connection: SQLite connection
            directory: path to the SiteOne directory
            site_id: ID for the site
            index_state: IndexState object for tracking progress
        """

        if not directory.exists() or not directory.is_dir():
            logger.error(f"Directory not found or not a directory: {directory}")
            return

        if index_state is not None:
            index_state.set_status(IndexStatus.INDEXING)

        log_data, log_http_error_data = self._extract_log_metadata(directory)

        file_paths = []
        for root, _, files in os.walk(directory):
            for filename in files:
                if filename == "robots.txt" or (filename.startswith("output.") and filename.endswith(".txt")):
                    continue
                file_paths.append(Path(root) / filename)

        processed_urls = set()

        with closing(connection.cursor()) as cursor:
            for i in range(0, len(file_paths), INDEXED_BATCH_SIZE):
                if index_state is not None and index_state.is_timeout():
                    index_state.set_status(IndexStatus.PARTIAL)
                    return

                batch_paths = file_paths[i:i+INDEXED_BATCH_SIZE]
                batch_insert_crawled: list[ResourceResult] = []
                file_contents = BaseManager.read_files(batch_paths)
                for file_path in batch_paths:
                    try:
                        result: ResourceResult | None = self._prepare_siteone_record(file_path,
                                site_id, directory, log_data, file_contents.get(file_path))
                        if result and result.url not in processed_urls:
                            batch_insert_crawled.append(result)
                            processed_urls.add(result.url)
                            if index_state is not None:
                                index_state.increment_processed()
                    except Exception as ex:
                        logger.error(f"Error processing file {file_path}: {ex}")

                self._execute_batch_insert(connection, cursor, batch_insert_crawled)

            # HTTP errors not already processed
            batch_insert_errors: list[ResourceResult] = []
            for url, meta in log_http_error_data.items():
                if url not in processed_urls:
                    size = meta.get("size", 0)
                    result = ResourceResult(
                        id=BaseManager.string_to_id(url),
                        site=site_id,
                        url=url,
                        type=ResourceResultType.OTHER,
                        status=meta["status"],
                        headers=BaseManager.get_basic_headers(size, ResourceResultType.OTHER, file_path),
                        content="",     # no content
                        size=size,      # size from log
                        time=meta["time"]
                    )
                    batch_insert_errors.append(result)

                    if index_state is not None:
                        index_state.increment_processed()

                    # errors in batches too
                    if len(batch_insert_errors) >= INDEXED_BATCH_SIZE:
                        self._execute_batch_insert(connection, cursor, batch_insert_errors)

            # insert any remaining error records
            if batch_insert_errors:
                self._execute_batch_insert(connection, cursor, batch_insert_errors)

            if index_state is not None and index_state.status == IndexStatus.INDEXING:
                index_state.set_status(IndexStatus.COMPLETE)

    def _prepare_siteone_record(self, file_path: Path, site_id: int, base_dir: Path,
                            log_data: dict, content: str = None) -> ResourceResult | None:
        """
        Prepare a record for batch insertion from a SiteOne file.

        Args:
            file_path: path to the file
            site_id: id for the site
            base_dir: base directory for the capture
            log_data: dictionary of metadata from logs keyed by URL
            content: optional pre-loaded file content

        Returns:
            Tuple of (record tuple, URL) or None if processing fails
        """
        try:
            # generate relative url path from file path (similar to wget)
            relative_path = file_path.relative_to(base_dir)
            url = f"{INDEXED_RESOURCE_DEFAULT_PROTOCOL}{base_dir.name}/{str(relative_path).replace(os.sep, '/')}"

            if file_path.is_file():
                file_stat = file_path.stat()
                file_size = file_stat.st_size
                file_created = datetime.fromtimestamp(file_stat.st_ctime, tz=timezone.utc)
                file_modified = datetime.fromtimestamp(file_stat.st_mtime, tz=timezone.utc)
            else:
                file_created = None
                file_modified = None
                file_size = 0

            decruftified_path = BaseManager.decruft_path(str(file_path))
            extension = Path(decruftified_path).suffix.lower()
            wget_static_pattern = re.compile(r"\.[0-9a-f]{8,}\.")

            # look up metadata from log if available, otherwise use defaults
            metadata = None
            wget_aliases = list(set([
                url,                                   # exact match first
                re.sub(wget_static_pattern, ".", url), # static pattern
                url.replace(".html", ""),              # file without extension (redirects)
                url.replace(".html", "/"),             # directory style (targets)
                url.replace("index.html", ""),         # index removal
            ]))

            for wget_alias in wget_aliases:
                metadata = log_data.get(wget_alias, None)
                if metadata is not None:
                    break

            if metadata is not None:
                # preventing duplicate html pages ./appstat.html and ./appstat/index.html
                # prefer index.html (actual content) over redirect stubs
                canonical_url = None
                # Sort aliases to prefer index.html files over redirect stubs
                sorted_aliases = sorted([alias for alias in wget_aliases if log_data.get(alias) == metadata],
                                    key=lambda x: (not x.endswith('index.html'), x))

                if sorted_aliases:
                    canonical_url = sorted_aliases[0]  # Take the preferred one
                    url = canonical_url
            else:
                metadata = {}

            status_code = metadata.get("status", 200)
            response_time = metadata.get("time", 0)
            log_type = metadata.get("type", "").lower()

            if log_type:
                # no type for redirects, but more often than not
                # redirection to another page
                resource_type = SITEONE_LOG_TYPE_MAPPING.get(log_type,
                        INDEXED_TYPE_MAPPING.get(extension, ResourceResultType.OTHER))
            else:
                # fallback to extension-based mapping
                resource_type = INDEXED_TYPE_MAPPING.get(extension, ResourceResultType.OTHER)

            file_content = content
            if file_content is None:
                file_content = BaseManager.read_file_contents(file_path, resource_type)

            # skip redirect stub files left in SiteOne archive (duplicate, wait for real content)
            if status_code == 200 and file_content and '<meta http-equiv="refresh" content="0' in file_content:
                return None

            record = ResourceResult(
                id=BaseManager.string_to_id(url),
                site=site_id,
                created=file_created,
                modified=file_modified,
                url=url,
                type=resource_type,
                status=status_code,
                headers=BaseManager.get_basic_headers(file_size, resource_type, file_path),
                content=file_content,
                size=file_size,
                time=response_time  # possibly from log
            )
            return record
        except Exception as ex:
            logger.error(f"Error preparing record for file {file_path}: {ex}")
            return None

manager: SiteOneManager = SiteOneManager()

def get_sites(
        datasrc: Path,
        ids: list[int] | None = None,
        fields: list[str] | None = None
    ) -> list[SiteResult]:
    """
    List site directories in the datasrc directory as sites.

    Args:
        datasrc: path to the directory containing site subdirectories
        ids: optional list of site IDs to filter by
        fields: optional list of fields to include in the response

    Returns:
        List of SiteResult objects, one for each site directory

    Notes:
        Returns an empty list if the datasrc directory doesn't exist.
    """
    return manager.get_sites_for_directories(datasrc, ids, fields)

def get_resources(
    datasrc: Path,
    sites: list[int] | None = None,
    query: str = "",
    fields: list[str] | None = None,
    sort: str | None = None,
    limit: int = RESOURCES_LIMIT_DEFAULT,
    offset: int = 0,
) -> tuple[list[ResourceResult], int, IndexState]:
    """
    Get resources from wget directories using in-memory SQLite.

    Args:
        datasrc: path to the directory containing wget captures
        sites: optional list of site IDs to filter by
        query: search query string
        fields: optional list of fields to include in response
        sort: sort order for results
        limit: maximum number of results to return
        offset: number of results to skip for pagination

    Returns:
        Tuple of (list of ResourceResult objects, total count)
    """
    sites_results: list[SiteResult] = get_sites(datasrc=datasrc, ids=sites)
    assert sites_results, "At least one site is required to search"
    site_paths = [site.path for site in sites_results]
    sites_group = SitesGroup(datasrc, sites, site_paths)
    return manager.get_resources_for_sites_group(sites_group, query, fields, sort, limit, offset)



import unittest
from mcp_server_webcrawl.utils.search import SearchQueryParser, SearchSubquery

class TestSearchQueryParser(unittest.TestCase):

    def setUp(self):
        """
        Set up a parser instance for each test
        """
        self.parser = SearchQueryParser()

    def test_simple_term(self):
        """
        Simple single term search
        """
        query = "hello"
        result: SearchSubquery= self.parser.parse(query)
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].field, None)
        self.assertEqual(result[0].value, "hello")
        self.assertEqual(result[0].type, "term")
        self.assertEqual(result[0].operator, None)

    def test_quoted_phrase(self):
        """
        Quoted phrase search
        """
        query = '"hello world"'
        result: SearchSubquery= self.parser.parse(query)
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].field, None)
        self.assertEqual(result[0].value, "hello world")
        self.assertEqual(result[0].type, "phrase")

    def test_wildcard_term(self):
        """
        Wildcard term search
        """
        query = "search*"
        result: SearchSubquery= self.parser.parse(query)
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].field, None)
        self.assertEqual(result[0].value, "search")
        self.assertEqual(result[0].type, "wildcard")

    def test_field_term(self):
        """
        Field-specific term search
        """
        query = "url:example.com"
        result: SearchSubquery= self.parser.parse(query)
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].field, "url")
        self.assertEqual(result[0].value, "example.com")
        self.assertEqual(result[0].type, "term")

    def test_field_numeric(self):
        """
        Field with numeric value
        """
        query = "status:404"
        result: SearchSubquery= self.parser.parse(query)
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].field, "status")
        self.assertEqual(result[0].value, 404)
        self.assertEqual(result[0].type, "term")

    def test_field_quoted(self):
        """
        Field with quoted value
        """
        query = 'content:"hello world"'
        result: SearchSubquery= self.parser.parse(query)
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].field, "content")
        self.assertEqual(result[0].value, "hello world")
        self.assertEqual(result[0].type, "phrase")

    def test_field_wildcard(self):
        """
        Field with wildcard value
        """
        query = "url:example*"
        result: SearchSubquery= self.parser.parse(query)
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].field, "url")
        self.assertEqual(result[0].value, "example")
        self.assertEqual(result[0].type, "wildcard")

    def test_simple_and(self):
        """
        Simple AND query
        """
        query = "hello AND world"
        result: SearchSubquery= self.parser.parse(query)
        self.assertEqual(len(result), 2)
        self.assertEqual(result[0].value, "hello")
        self.assertEqual(result[0].operator, "AND")
        self.assertEqual(result[1].value, "world")
        self.assertEqual(result[1].operator, None)

    def test_simple_or(self):
        """
        Simple OR query
        """
        query = "hello OR world"
        result: SearchSubquery= self.parser.parse(query)
        self.assertEqual(len(result), 2)
        self.assertEqual(result[0].value, "hello")
        self.assertEqual(result[0].operator, "OR")
        self.assertEqual(result[1].value, "world")
        self.assertEqual(result[1].operator, None)

    def test_simple_not(self):
        """
        Simple NOT query
        """
        query = "NOT hello"
        result: SearchSubquery= self.parser.parse(query)
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].value, "hello")
        self.assertTrue('NOT' in result[0].modifiers)

    def test_and_with_fields(self):
        """
        AND with field specifiers
        """
        query = "content:hello AND url:example.com"
        result: SearchSubquery= self.parser.parse(query)
        self.assertEqual(len(result), 2)
        self.assertEqual(result[0].field, "content")
        self.assertEqual(result[0].operator, "AND")
        self.assertEqual(result[1].field, "url")

    def test_or_with_fields(self):
        """
        OR with field specifiers
        """
        query = "status:404 OR status:500"
        result: SearchSubquery= self.parser.parse(query)
        self.assertEqual(len(result), 2)
        self.assertEqual(result[0].field, "status")
        self.assertEqual(result[0].value, 404)
        self.assertEqual(result[0].operator, "OR")
        self.assertEqual(result[1].field, "status")
        self.assertEqual(result[1].value, 500)

    def test_not_with_field(self):
        """
        NOT with field specifier
        """
        query = "NOT status:404"
        result: SearchSubquery= self.parser.parse(query)
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].field, "status")
        self.assertEqual(result[0].value, 404)
        self.assertTrue('NOT' in result[0].modifiers)

    def test_simple_parentheses(self):
        """
        Simple expression with parentheses
        """
        query = "(hello AND world)"
        result: SearchSubquery= self.parser.parse(query)
        self.assertEqual(len(result), 2)
        self.assertEqual(result[0].value, "hello")
        self.assertEqual(result[0].operator, "AND")
        self.assertEqual(result[1].value, "world")
        self.assertEqual(result[1].operator, None)

    def test_complex_parentheses(self):
        """
        Complex expression with nested parentheses
        """
        query = "(hello AND (world OR planet))"
        result: SearchSubquery= self.parser.parse(query)
        self.assertEqual(len(result), 3)
        self.assertEqual(result[0].value, "hello")
        self.assertEqual(result[0].operator, "AND")
        self.assertEqual(result[1].value, "world")
        self.assertEqual(result[1].operator, "OR")
        self.assertEqual(result[2].value, "planet")
        self.assertEqual(result[2].operator, None)

    def test_mixed_operators(self):
        """
        Query with mixed operators
        """
        query = "hello AND world OR planet"
        result: SearchSubquery= self.parser.parse(query)
        self.assertEqual(len(result), 3)
        self.assertEqual(result[0].value, "hello")
        self.assertEqual(result[0].operator, "AND")
        self.assertEqual(result[1].value, "world")
        self.assertEqual(result[1].operator, "OR")
        self.assertEqual(result[2].value, "planet")
        self.assertEqual(result[2].operator, None)

    def test_mixed_with_parentheses(self):
        """
        Mixed operators with parentheses for precedence
        """
        query = "hello AND (world OR planet)"
        result: SearchSubquery= self.parser.parse(query)
        self.assertEqual(len(result), 3)
        self.assertEqual(result[0].value, "hello")
        self.assertEqual(result[0].operator, "AND")
        self.assertEqual(result[1].value, "world")
        self.assertEqual(result[1].operator, "OR")
        self.assertEqual(result[2].value, "planet")
        self.assertEqual(result[2].operator, None)

    def test_complex_nested_query(self):
        """
        Complex nested query with multiple operators
        """
        query = '(content:"error message" AND (status:404 OR status:500)) AND NOT url:example.com'
        result: SearchSubquery= self.parser.parse(query)
        self.assertEqual(len(result), 4)
        self.assertEqual(result[0].field, "content")
        self.assertEqual(result[0].value, "error message")
        self.assertEqual(result[0].operator, "AND")
        self.assertEqual(result[1].field, "status")
        self.assertEqual(result[1].value, 404)
        self.assertEqual(result[1].operator, "OR")
        self.assertEqual(result[2].field, "status")
        self.assertEqual(result[2].value, 500)
        self.assertEqual(result[2].operator, "NOT")
        self.assertEqual(result[3].field, "url")
        self.assertEqual(result[3].value, "example.com")
        self.assertEqual(result[3].operator, None)

    def test_all_features_combined(self):
        """
        Comprehensive test with all features combined
        """
        query = 'content:"critical error" AND (status:500 OR type:html) AND NOT url:example* AND size:1024'
        result: SearchSubquery= self.parser.parse(query)
        self.assertEqual(len(result), 5)
        self.assertEqual(result[0].field, "content")
        self.assertEqual(result[0].value, "critical error")
        self.assertEqual(result[0].type, "phrase")
        self.assertEqual(result[0].operator, "AND")
        self.assertEqual(result[1].field, "status")
        self.assertEqual(result[1].value, 500)
        self.assertEqual(result[1].operator, "OR")
        self.assertEqual(result[2].field, "type")
        self.assertEqual(result[2].value, "html")
        self.assertEqual(result[2].operator, "AND")
        self.assertEqual(result[3].field, "url")
        self.assertEqual(result[3].value, "example")
        self.assertEqual(result[3].type, "wildcard")
        self.assertTrue('NOT' in result[3].modifiers)
        self.assertEqual(result[3].operator, "AND")
        self.assertEqual(result[4].field, "size")
        self.assertEqual(result[4].value, 1024)
        self.assertEqual(result[4].operator, None)

    def test_to_sqlite_fts(self):
        """
        Test conversion to SQLite FTS format
        """
        query = 'content:"error" AND status:404'
        result: SearchSubquery= self.parser.parse(query)

        query_parts, params = self.parser.to_sqlite_fts(result)

        self.assertEqual(len(query_parts), 3)
        self.assertEqual(query_parts[0], "ResourcesFullText.Content MATCH :query0")
        self.assertEqual(query_parts[1], "AND")
        self.assertEqual(query_parts[2], "Resources.Status = :query1")

        self.assertEqual(len(params), 2)
        self.assertEqual(params["query0"], '"error"')
        self.assertEqual(params["query1"], 404)

    def test_operator_assignment_bug(self):
        """
        Test that exposes the double operator assignment bug.
        Query: "term1 AND term2 OR term3" should create:
        [term1(op=AND), term2(op=OR), term3(op=None)]

        Were the bug present, term3 would incorrectly get operator == OR
        """
        from mcp_server_webcrawl.utils.parser import SearchLexer, SearchParser

        lexer = SearchLexer()
        parser = SearchParser(lexer)
        query = "term1 AND term2 OR term3"
        result = parser.parser.parse(query, lexer=lexer.lexer)
        self.assertEqual(len(result), 3)
        self.assertEqual(result[0].value, "term1")
        self.assertEqual(result[0].operator, "AND")
        self.assertEqual(result[1].value, "term2")
        self.assertEqual(result[1].operator, "OR")
        self.assertEqual(result[2].value, "term3")
        self.assertEqual(result[2].operator, None)



import curses

from enum import Enum, auto
from typing import NamedTuple, Optional, Tuple

from mcp_server_webcrawl.crawlers import VALID_CRAWLER_CHOICES

SITE_COLUMN_WIDTH = 18

LAYOUT_GRID_COLUMN_SPACING = 2
LAYOUT_CONSTRAINED_SITES_PER_COLUMN = 3
LAYOUT_SITES_GRID_OFFSET = 6

DEFAULT_GROUP_WIDTH = 12

INPUT_BOX_BRACKET_WIDTH = 2
CURSOR_SCROLL_THRESHOLD = 5

class DocumentMode(Enum):
    MARKDOWN = auto()
    RAW = auto()
    HEADERS = auto()

class NavigationDirection(Enum):
    UP = auto()
    DOWN = auto()
    LEFT = auto()
    RIGHT = auto()

class ScrollDirection(Enum):
    UP = auto()
    DOWN = auto()

class SearchFilterType(Enum):
    ANY = 0
    PAGES = 1

class ThemeDefinition(Enum):
    # https://www.ditig.com/256-colors-cheat-sheet
    DOCUMENT_MODE = (1, curses.COLOR_BLUE, 51)
    HEADER_ACTIVE = (2, curses.COLOR_WHITE, 17)
    HEADER_INACTIVE = (3, curses.COLOR_WHITE, 233)
    HEADER_OUTER = (4, curses.COLOR_WHITE, 235)
    HELP_LINK = (5, curses.COLOR_WHITE, 27)
    HTTP_ERROR = (6, curses.COLOR_WHITE, 88)
    HTTP_WARN = (7, curses.COLOR_WHITE, 130)
    INACTIVE_QUERY = (8, 245, 237)
    SNIPPET_DEFAULT = (9, 243, curses.A_DIM)
    SNIPPET_HIGHLIGHT = (10, 232, 51)
    UI_ERROR = (11, curses.COLOR_WHITE, 88)

class UiFocusable(Enum):
    UNDEFINED = auto()
    SEARCH_FORM = auto()
    SEARCH_RESULTS = auto()

class UiState(Enum):
    UNDEFINED = auto()
    REQUIREMENTS = auto()
    SEARCH_INIT = auto()
    SEARCH_RESULTS = auto()
    DOCUMENT = auto()
    HELP = auto()

def safe_addstr(stdscr: curses.window, y: int, x: int, text: str, style: int = curses.A_NORMAL) -> None:
    """
    Safe addstr that handles screen edge errors.
    """
    try:
        stdscr.addstr(y, x, text, style)
    except curses.error:
        pass

class InputRadio:
    def __init__(self, group, name: str, label: str, index: int, states: list = None):
        """
        Radio input with 2-3 possible states (e.g., on/off or state1/state2/off)
        
        Args:
            group: The InputRadioGroup this radio belongs to
            name: The form radio group name
            label: The form radio label  
            index: The current state index
            states: List of InputRadioState objects defining each possible state            
        """
        # used like so states_radio = InputRadio(groupname, label, index, states=InputRadioGroup.get_filters())
        if states is None:
            states = []
        assert states, "states must be provided and non-empty"
        assert 0 <= index < len(states), f"index {index} out of range for {len(states)} states"

        self.name = name
        self.label = label
        self.index = index
        self._states = states
        self._group = group


    @property
    def current_state(self):
        """
        Get the current state
        """
        return self._states[self.index]

    @property
    def display_label(self) -> str:
        """
        Get the current display label
        """
        return self.current_state.label

    @property
    def value(self) -> str:
        """
        Get the current value
        """
        return self.current_state.value

    def next_state(self) -> None:
        """
        Cycle to the next state
        """
        # clear group for single-selection radios
        if self._group.name in ["filter", "site", "crawler"]:
            self._group.clear()

        if self._group.name == "sort":
            if self.index == 0:         # inactive " " -> active ascending "+"
                self._group.clear()
                self.index = 1
            elif self.index == 1:       # ascending "+" -> descending "-"
                self.index = 2
            elif self.index == 2:       # descending "-" -> ascending "+"
                self.index = 1
        else:
            # standard cycling for other radios
            self.index = (self.index + 1) % len(self._states)

    def render(self, stdscr: curses.window, y: int, x: int, field_index: int, max_width: int = None, focused: bool = False) -> None:
        """
        Render a single radio option.
        """

        radio_symbol = self.display_label
        display_text = self.label
        if max_width and len(display_text) > max_width:
            display_text = display_text[:max_width - 1] + "…"

        line = f"({radio_symbol}) {display_text}"
        style = curses.A_REVERSE if focused else curses.A_NORMAL

        try:
            safe_addstr(stdscr, y, x, line, style)
        except curses.error:
            pass  # screen edge

    def set_state(self, index: int) -> None:
        """
        Set the current state by index
        """
        if 0 <= index < len(self._states):
            self.index = index
        else:
            raise IndexError(f"State index {index} out of range")

    def set_states(self, states: list) -> None:
        """
        Set the current state by index
        """
        self._states = states

    def __str__(self) -> str:
        return f"{self.label}: {self.display_label} ({self.value})"

class InputRadioGroup:
    """
    Radio group with navigation and layout management capabilities.
    """
    def __init__(self, name: str, sites: list = None):
        """
        Radio input group with layout and navigation support.
        
        Args:
            name: The form radio group name ("filter", "sort", "site", or "crawler")
            sites: List of SiteResult objects, required only for "site" group type
        """
        sites = sites if sites is not None else []
        self.name: str = name
        self.label: str = name
        self.__selected_index: int = 0

        # layout configuration
        self.__available_width: int = 0
        self.__available_height: int = 0
        self.__is_constrained: bool = False
        self.__sites_per_column: int = 0
        self.__max_columns: int = 0

        self.radios: list[InputRadio] = []

        group_config = {
            "filter": (self.__load_filters, "Filter:"),
            "site": (lambda: self.__load_sites(sites), "Sites:"),
            "sort": (self.__load_sorts, "Sorts:"),
            "crawler": (self.__load_crawlers, "Crawlers:"),
        }

        if self.name in group_config:
            data_loader, label = group_config[self.name]
            self.label = label
            data_loader()
        else:
            raise Exception(f"Unsupported radio option: {self.name}")

        if self.radios:
            self.radios[0].next_state()

    @property
    def value(self) -> str:
        for radio in self.radios:
            if radio.value == "on" or radio.display_label in ["+", "-"]:  # selected state
                if self.name == "filter":
                    return "html" if radio.label == "HTML" else ""
                elif self.name == "sort":
                    if radio.display_label == "+":
                        return f"+{radio.label}"
                    elif radio.display_label == "-":
                        return f"-{radio.label}"
                    return ""
                elif self.name == "site":
                    return radio.label  # or site ID/URL however you want to identify it
                elif self.name == "crawler":
                    return radio.label
        return ""

    def calculate_group_width(self) -> int:
        """
        Calculate the display width needed for a radio group.
        """
        if not self.radios:
            return DEFAULT_GROUP_WIDTH
        return max(len(radio.label) for radio in self.radios)

    def clear(self) -> None:
        for r in self.radios:
            r.index = 0

    def set_layout_constraints(self, available_width: int, available_height: int, is_constrained: bool = False) -> None:
        """
        Set layout constraints for grid-based groups (like sites).
        
        Args:
            available_width: Available horizontal space
            available_height: Available vertical space  
            is_constrained: Whether layout is constrained (affects sites per column)
        """
        self.__available_width = available_width
        self.__available_height = available_height
        self.__is_constrained = is_constrained

        if self.name == "site":
            self.__calculate_grid_layout()

    def get_grid_position(self, radio_index: int) -> Tuple[int, int]:
        """
        Convert linear radio index to grid position.
        Only applies to site groups; other groups return (radio_index, 0).
        
        Args:
            radio_index: Linear index in radios list
            
        Returns:
            tuple: (row, column) position in grid layout
        """
        if self.name != "site" or self.__sites_per_column == 0:
            return (radio_index, 0)

        row = radio_index % self.__sites_per_column
        col = radio_index // self.__sites_per_column
        return (row, col)

    def get_index_from_grid(self, row: int, col: int) -> Optional[int]:
        """
        Convert grid position to linear radio index.
        Only works for site groups; returns None for other group types.
    
        Args:
            row: Row in grid (0-based)
            col: Column in grid (0-based)
    
        Returns:
            Linear index if position exists within grid bounds, None otherwise
        """
        if self.name != "site":
            return row if 0 <= row < len(self.radios) else None
#
        if self.__sites_per_column == 0:
            return None
#
        radio_index = col * self.__sites_per_column + row
        if (0 <= radio_index < len(self.radios) and
            radio_index < self.__sites_per_column * self.__max_columns):
            return radio_index
        return None

    def navigate_left(self, current_radio_index: int) -> Optional[int]:
        """
        Navigate left within this group's layout.
        
        Args:
            current_radio_index: Current position in radios list
            
        Returns:
            New radio index if navigation successful, None if should exit group
        """
        if self.name != "site":
            # don't support horizontal navigation
            return None

        current_row, current_col = self.get_grid_position(current_radio_index)

        if current_col > 0:
            # to previous column, same row
            return self.get_index_from_grid(current_row, current_col - 1)
        else:
            # at leftmost column, signal exit to parent
            return None

    def navigate_right(self, current_radio_index: int) -> Optional[int]:
        """
        Navigate right within this group's layout.
        
        Args:
            current_radio_index: Current position in radios list
            
        Returns:
            New radio index if navigation successful, None if should exit group
        """
        if self.name != "site":
            # don't support horizontal navigation
            return None

        current_row, current_col = self.get_grid_position(current_radio_index)
        new_index = self.get_index_from_grid(current_row, current_col + 1)
        return new_index  # if invalid/out of bounds

    def navigate_to_row(self, target_row: int, from_column: int = 0) -> Optional[int]:
        """
        Navigate to a specific row from an external column position.
        """
        if self.name != "site":
            return target_row if 0 <= target_row < len(self.radios) else None

        if self.__sites_per_column == 0:
            return target_row if 0 <= target_row < len(self.radios) else None

        return self.get_index_from_grid(target_row, from_column)

    def get_row_from_index(self, radio_index: int) -> int:
        """
        Get the row number for navigation between groups.
        
        Args:
            radio_index: Linear index in radios list
            
        Returns:
            Row number for inter-group navigation
        """
        if self.name != "site":
            return radio_index

        row, _ = self.get_grid_position(radio_index)
        return row

    def __calculate_grid_layout(self) -> None:
        """
        Calculate grid layout parameters for sites group.
        """
        if self.name != "site":
            return

        self.__sites_per_column = (LAYOUT_CONSTRAINED_SITES_PER_COLUMN if self.__is_constrained
                                 else min(self.__available_height - LAYOUT_SITES_GRID_OFFSET, len(self.radios)))

        if self.__available_width > SITE_COLUMN_WIDTH:
            self.__max_columns = max(1, self.__available_width // (SITE_COLUMN_WIDTH + LAYOUT_GRID_COLUMN_SPACING))
        else:
            self.__max_columns = 1

    def __display_url(self, url: str) -> str:
        return url.split("://")[-1].rstrip("/")

    def __get_on_off_state(self) -> list:
        return [
            InputRadioState(" ", ""),
            InputRadioState("●", "on")
        ]

    def __load_crawlers(self) -> None:
        # "archivebox", "httrack", "interrobot", "katana", "siteone", "warc", "wget"
        self.radios = [
            InputRadio(self, "crawler", label, 0, self.__get_on_off_state())
            for label in VALID_CRAWLER_CHOICES
        ]

    def __load_filters(self) -> None:
        self.radios = [
            InputRadio(self, "filter", "HTML", 0, self.__get_on_off_state()),
            InputRadio(self, "filter", "any", 0, self.__get_on_off_state())
        ]

    def __load_sites(self, sites: list) -> None:
        site_labels = [self.__display_url(s.urls[0]) if s.urls else "unknown" for s in sites]
        self.radios = [
            InputRadio(self, "site", label, 0, self.__get_on_off_state())
            for label in site_labels
        ]

    def __load_sorts(self) -> None:
        sort_states = [
            InputRadioState(" ", ""),
            InputRadioState("+", "+"),
            InputRadioState("-", "-")
        ]

        sort_labels = ["URL", "status", "size"]

        self.radios = [
            InputRadio(self, "sort", label, 0, sort_states.copy())
            for label in sort_labels
        ]

class InputRadioState(NamedTuple):
    label: str  # "●", " ", "-", "+"
    value: str  # "", "+url", "-sort"

class InputText:
    """
    A reusable text input field with cursor management, rendering, and input handling.
    Consolidates the common text input functionality used across the application.
    """

    def __init__(self, initial_value: str = "", max_length: int = None, label: str = ""):
        """
        Initialize the text input field.
        
        Args:
            initial_value: Starting text value
            max_length: Maximum allowed text length (None for unlimited)
            label: Display label for the field
        """
        self.value: str = initial_value
        self.cursor_pos: int = len(initial_value)
        self.max_length: int = max_length
        self.label: str = label

        self._last_display_cache: Optional[tuple] = None
        self._last_value_hash: int = 0

    def backspace(self) -> None:
        """
        Remove the character before the cursor.
        """
        if self.cursor_pos > 0:
            self.value = self.value[:self.cursor_pos - 1] + self.value[self.cursor_pos:]
            self.cursor_pos -= 1

    def clear(self) -> None:
        """
        Clear all text and reset cursor.
        """
        self.value = ""
        self.cursor_pos = 0

    def delete(self) -> None:
        """
        Remove the character at the cursor position.
        """
        if self.cursor_pos < len(self.value):
            self.value = self.value[:self.cursor_pos] + self.value[self.cursor_pos + 1:]

    def end(self) -> None:
        """
        Move cursor to the end of the text.
        """
        self.cursor_pos = len(self.value)

    def handle_input(self, key: int) -> bool:
        """
        Handle keyboard input for the text field.
        
        Args:
            key: The curses key code
            
        Returns:
            bool: True if the input was handled, False otherwise
        """
        handlers: dict[int, callable] = {
            curses.KEY_LEFT: self.move_cursor_left,
            curses.KEY_RIGHT: self.move_cursor_right,
            curses.KEY_HOME: self.home,
            curses.KEY_END: self.end,
            curses.KEY_BACKSPACE: self.backspace,
            127: self.backspace,  # alternative backspace
            8: self.backspace,    # alternative backspace
            curses.KEY_DC: self.delete,
        }

        handler = handlers.get(key)
        if handler:
            handler()
            return True

        if 32 <= key <= 126:  # printable characters
            char: str = chr(key)
            self.insert_char(char)
            return True

        return False

    def home(self) -> None:
        """
        Move cursor to the beginning of the text.
        """
        self.cursor_pos = 0

    def insert_char(self, char: str) -> None:
        """
        Insert a character at the current cursor position.
        """
        sanitized = self.__sanitize_input(char)
        if sanitized is None:
            return

        if self.max_length is not None and len(self.value) >= self.max_length:
            return

        self.value = self.value[:self.cursor_pos] + char + self.value[self.cursor_pos:]
        self.cursor_pos += 1

    def is_empty(self) -> bool:
        """
        Check if the text field is empty.
        """
        return len(self.value.strip()) == 0

    def move_cursor_left(self) -> None:
        """
        Move cursor one position to the left.
        """
        if self.cursor_pos > 0:
            self.cursor_pos -= 1

    def move_cursor_right(self) -> None:
        """
        Move cursor one position to the right.
        """
        if self.cursor_pos < len(self.value):
            self.cursor_pos += 1

    def render(self, stdscr: curses.window, y: int, x: int, width: int,
               focused: bool = False, style: int = None) -> None:
        """
        Render the text input field with box, text, and cursor.
        
        Args:
            stdscr: The curses window
            y: Y position to render at
            x: X position to render at  
            width: Total width of the input box
            focused: Whether this field has focus (shows cursor)
            style: Curses style attributes to apply
        """

        # account for [ ] brackets
        inner_width = max(1, width - INPUT_BOX_BRACKET_WIDTH)
        display_text, display_cursor_pos = self.__calculate_display_text_and_cursor(inner_width)
        box_content = f"[{display_text.ljust(inner_width)}]"
        if style is None:
            style = curses.A_REVERSE if focused else curses.A_NORMAL

        safe_addstr(stdscr, y, x, box_content, style)
        if focused:
            self.__render_cursor(stdscr, y, x, display_text, display_cursor_pos, inner_width)

    def set_value(self, new_value: str) -> None:
        """
        Set the text value and adjust cursor if needed.
        """
        self.value = new_value
        # cursor doesn't go beyond text length
        self.cursor_pos = min(self.cursor_pos, len(self.value))

    def __sanitize_input(self, char: str) -> Optional[str]:
        """
        Sanitize input character, return None if should be rejected
        """
        # strip control characters
        if ord(char) < 32 or ord(char) == 127:
            return None
        # add more checks here as needed
        return char

    def __render_cursor(self, stdscr: curses.window, y: int, x: int,
                      display_text: str, display_cursor_pos: int, inner_width: int) -> None:
        """
        Render the cursor at the appropriate position.
        
        Args:
            stdscr: The curses window
            y: Y position of the input box
            x: X position of the input box
            display_text: The currently displayed text
            display_cursor_pos: Where the cursor appears in the displayed text
            inner_width: Available width inside the box
        """

        try:
            if display_cursor_pos < len(display_text) and display_cursor_pos < inner_width:
                cursor_x = x + 1 + display_cursor_pos
                # highlight the character under cursor instead of just reversing
                char_under_cursor = display_text[display_cursor_pos]
                safe_addstr(stdscr, y, cursor_x, char_under_cursor, curses.A_REVERSE | curses.A_BOLD)
            elif display_cursor_pos >= 0 and x + 1 + display_cursor_pos < x + 1 + inner_width:
                # cursor at end - underscore
                cursor_x = x + 1 + display_cursor_pos
                safe_addstr(stdscr, y, cursor_x, '_', curses.A_REVERSE | curses.A_BOLD)
        except curses.error:
            pass


    def __calculate_display_text_and_cursor(self, inner_width: int) -> tuple[str, int]:
        """
        Calculate what portion of text to display and where the cursor should appear.
        Handles horizontal scrolling for long text.
        
        Args:
            inner_width: Available width inside the input box
            
        Returns:
            tuple: (display_text, display_cursor_position)
        """
        current_hash = hash((self.value, self.cursor_pos, inner_width))
        if current_hash == self._last_value_hash and self._last_display_cache:
            return self._last_display_cache

        if len(self.value) <= inner_width:
            # text fits entirely
            return self.value, self.cursor_pos

        # text is longer than available space, scroll
        if self.cursor_pos >= inner_width - CURSOR_SCROLL_THRESHOLD:
            start_pos = max(0, len(self.value) - inner_width)
            display_text = self.value[start_pos:]
            display_cursor_pos = self.cursor_pos - start_pos
        else:
            display_text = self.value[:inner_width]
            display_cursor_pos = min(self.cursor_pos, inner_width)

        return display_text, display_cursor_pos

class ViewBounds:
    def __init__(self, x: int = 0, y: int = 0, width: int = 0, height: int = 0):
        self.x = x
        self.y = y
        self.width = width
        self.height = height
